"""
Classes for handling locking and ownerships.
"""
import os
import os.path
import fcntl
from ruruki import interfaces


class Lock(interfaces.ILock):
    """
    Base locking class.

    See :class:`~.ILock` for doco.
    """

    def __init__(self):
        self._locked = False

    @property
    def locked(self):
        """
        Return the status of the lock.

        :returns: True if the lock is acquired.
        :rtype: :class:`bool`
        """
        return self._locked

    def acquire(self):
        if self._locked is True:
            raise interfaces.AcquireError(
                "Failed to acquire lock."
            )
        self._locked = True

    def release(self):
        if self.locked is False:
            raise interfaces.ReleaseError(
                "Failed releasing lock."
            )
        self._locked = False

    def __del__(self):
        if self.locked is True:
            self.release()

    def __enter__(self):
        lock = self
        lock.acquire()
        return lock

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()


class FileLock(Lock):
    """
    File based locking.

    :param filename: Filename to create a lock on.
    :type filename: :class:`str`
    """

    def __init__(self, filename):
        super(FileLock, self).__init__()
        self.filename = filename
        self._fd = None

    def acquire(self):
        self._fd = open(self.filename, "w")

        try:
            fcntl.flock(self._fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise interfaces.AcquireError(
                "Failed acquiring a lock on {0!r}.".format(self.filename)
            )
        super(FileLock, self).acquire()

    def _close_file(self):
        if self._fd is not None:
            self._fd.close()

    def release(self):
        self._close_file()
        super(FileLock, self).release()

    def __del__(self):
        self._close_file()
        super(FileLock, self).__del__()


class DirectoryLock(Lock):
    """
    Directory based locking.

    :param path: Path that you are locking.
    :type path: :class:`str`
    """

    def __init__(self, path):
        super(DirectoryLock, self).__init__()
        self.path = path
        self.filename = os.path.join(path, ".lock")
        self._filelock = FileLock(self.filename)

    @property
    def locked(self):
        """
        Return the status of the lock.

        :returns: True if the lock is acquired.
        :rtype: :class:`bool`
        """
        return self._filelock.locked

    def acquire(self):
        try:
            self._filelock.acquire()
        except interfaces.AcquireError:
            raise interfaces.AcquireError(
                "Failed acquiring a lock for path {0!r}.".format(self.path)
            )

    def release(self):
        try:
            self._filelock.release()
        except interfaces.ReleaseError:
            raise interfaces.ReleaseError(
                "Failed releasing the lock for path {0!r}.".format(self.path)
            )
        finally:
            if os.path.isfile(self.filename):
                os.remove(self.filename)



# pylint: disable=missing-docstring
import os.path
import tempfile
import unittest2
from ruruki import interfaces
from ruruki import locks


class TestBaseLock(unittest2.TestCase):
    def setUp(self):
        self.lock = locks.Lock()

    def test_acquire(self):
        self.lock.acquire()
        self.assertEqual(self.lock.locked, True)

    def test_acquire_already_locked(self):
        self.lock.acquire()
        self.assertRaises(
            interfaces.AcquireError,
            self.lock.acquire
        )

    def test_release(self):
        self.lock.acquire()
        self.lock.release()
        self.assertEqual(self.lock.locked, False)

    def test_release_not_locked(self):
        self.assertRaises(
            interfaces.ReleaseError,
            self.lock.release
        )

    def test_delete_releases_lock(self):
        self.lock.acquire()
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)

    def test_delete_not_locked(self):
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)

    def test_enter_contex_manager(self):
        with locks.Lock() as lock:
            self.assertEqual(lock.locked, True)

    def test_exit_contex_manager(self):
        context_manager = locks.Lock()
        lock = context_manager.__enter__()
        lock.__exit__(None, None, None)
        self.assertEqual(lock.locked, False)


class TestFileLock(unittest2.TestCase):
    def setUp(self):
        self.filename = tempfile.NamedTemporaryFile()
        self.lock = locks.FileLock(self.filename.name)

    def test_acquire(self):
        self.lock.acquire()
        self.assertEqual(self.lock.locked, True)

    def test_acquire_flock_error(self):
        filename = tempfile.NamedTemporaryFile()
        with locks.FileLock(filename.name):
            lock = locks.FileLock(filename.name)
            self.assertRaises(
                interfaces.AcquireError,
                lock.acquire
            )

    def test_acquire_already_locked(self):
        self.lock.acquire()
        self.assertRaises(
            interfaces.AcquireError,
            self.lock.acquire,
        )

    def test_release(self):
        self.lock.acquire()
        self.lock.release()
        self.assertEqual(self.lock.locked, False)

    def test_release_not_locked(self):
        self.assertRaises(
            interfaces.ReleaseError,
            self.lock.release,
        )

    def test_delete_releases_lock(self):
        self.lock.acquire()
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)

    def test_delete_not_locked(self):
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)

    def test_enter_contex_manager(self):
        with locks.FileLock(self.filename.name) as lock:
            self.assertEqual(lock.locked, True)

    def test_exit_contex_manager(self):
        context_manager = locks.FileLock(self.filename.name)
        lock = context_manager.__enter__()
        lock.__exit__(None, None, None)
        self.assertEqual(lock.locked, False)


class TestDirectoryLock(unittest2.TestCase):
    def setUp(self):
        self.path = tempfile.mkdtemp()
        self.lock = locks.DirectoryLock(self.path)

    def test_acquire(self):
        self.lock.acquire()
        self.assertEqual(self.lock.locked, True)
        self.assertEqual(
            os.path.isfile(self.lock.filename),
            True,
        )

    def test_acquire_already_locked(self):
        self.lock.acquire()
        self.assertRaises(
            interfaces.AcquireError,
            self.lock.acquire,
        )

    def test_release(self):
        self.lock.acquire()
        self.lock.release()
        self.assertEqual(self.lock.locked, False)
        self.assertEqual(
            os.path.isfile(self.lock.filename),
            False,
        )

    def test_release_not_locked(self):
        self.assertRaises(
            interfaces.ReleaseError,
            self.lock.release,
        )

    def test_release_file_not_found(self):
        self.lock.acquire()
        os.remove(self.lock.filename)
        self.lock.release()
        self.assertEqual(self.lock.locked, False)

    def test_delete_releases_lock(self):
        self.lock.acquire()
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)
        self.assertEqual(
            os.path.isfile(self.lock.filename),
            False,
        )

    def test_delete_not_locked(self):
        self.lock.__del__()
        self.assertEqual(self.lock.locked, False)
        self.assertEqual(
            os.path.isfile(self.lock.filename),
            False,
        )

    def test_enter_contex_manager(self):
        with locks.DirectoryLock(self.path) as lock:
            self.assertEqual(lock.locked, True)
            self.assertEqual(
                os.path.isfile(lock.filename),
                True,
            )

    def test_exit_contex_manager(self):
        context_manager = locks.DirectoryLock(self.path)
        lock = context_manager.__enter__()
        lock.__exit__(None, None, None)
        self.assertEqual(lock.locked, False)
        self.assertEqual(
            os.path.isfile(lock.filename),
            False,
        )



# Due to the nature of the pyparsing grammar expressions, linting is not
# really valid, so we are disbling it for most things.
# pylint: disable=invalid-name,expression-not-assigned
"""
Grammar expressions.

See https://s3.amazonaws.com/artifacts.opencypher.org/cypher.ebnf
"""
import pyparsing as pp


###########################################
##  COMMON PYPARSING GRAMMAR EXPRESSIONS ##
###########################################

# abc...
var = pp.Word(pp.alphas)("alphas")

# 0123...
nums = pp.Word(pp.nums)("nums").setParseAction(lambda s, l, t: int(t[0]))

# abc...0123...
varnums = pp.Word(pp.alphanums).setResultsName("alphanums")


# =
eq_operation = (
    pp.Word("=", exact=1)("operation").setParseAction(lambda x: "eq")
)


# !=
neq_operation = (
    pp.Word("!=", exact=2)("operation").setParseAction(lambda x: "neq")
)

# <
lt_operation = (
    pp.Word("<", exact=1)("operation").setParseAction(lambda x: "lt")
)


# >
gt_operation = (
    pp.Word(">", exact=1)("operation").setParseAction(lambda x: "gt")
)


#>=
gte_operation = (
    pp.Word(">=", exact=2)("operation").setParseAction(lambda x: "gte")
)


# <=
lte_operation = (
    pp.Word("<=", exact=2)("operation").setParseAction(lambda x: "lte")
)


# +
add_operation = pp.Word("+", exact=1)("operation")


# -
subtract_operation = pp.Word("-", exact=1)("operation")


# *
multiply_operation = pp.Word("*", exact=1)("operation")


# divide
divide_operation = pp.Word("/", exact=1)("operation")


# %
percent_operation = pp.Word("%", exact=1)("operation")


# ^
bitwise_or_operation = pp.Word("^", exact=1)("operation")

# =~
regex_operation = pp.Word("=~", exact=2)("operation")

# with or WITH
with_keyword = pp.CaselessKeyword("WITH")("with_keyword")


# IN or in
in_keyword = pp.CaselessKeyword("IN")("in_keyword")


# IS or is
is_keyword = pp.CaselessKeyword("IS")("is_keyword")


# NOT or not
not_keyword = pp.CaselessKeyword("NOT")("not_keyword")


# NNULL or null
null_keyword = pp.CaselessKeyword("NULL")("null_keyword")


# STARTS or starts
starts_keyword = pp.CaselessKeyword("STARTS")("starts_keyword")


# CONTAINS or contains
contains_keyword = pp.CaselessKeyword("CONTAINS")("contains_keyword")


# ENDS or ends
ends_keyword = pp.CaselessKeyword("ENDS")("ends_keyword")


# STARTS WITH
starts_with = (
    (starts_keyword + with_keyword)("string_matching").
    setParseAction(lambda s: "startswith")
)


# ENDS WITH
ends_with = (
    (ends_keyword + with_keyword)("string_matching").
    setParseAction(lambda e: "endswith")
)

# CONTAINS
contains = (
    contains_keyword("string_matching").
    setParseAction(lambda c: "contains")
)


# IS NOT or is not
is_not = (
    (is_keyword + not_keyword)("string_matching").
    setParseAction(lambda i: "is_not")
)

# IS NULL or is null
is_null = (
    (is_keyword + null_keyword)("type_matching").
    setParseAction(lambda i: "is_null")
)


# IS NOT NULL or is not null
is_not_null = (
    (is_keyword + not_keyword + null_keyword)("type_matching").
    setParseAction(lambda i: "is_not_null")
)

# True or true
true_keyword = (
    pp.CaselessKeyword("TRUE")("type_matching").
    setParseAction(lambda x: True)
)

# False or false
false_keyword = (
    pp.CaselessKeyword("FALSE")("type_matching").
    setParseAction(lambda x: False)
)


# ((C,O,U,N,T), '(', '*', ')')
count = (
    (
        pp.CaselessKeyword("COUNT") +
        pp.Suppress(pp.Word("(", exact=1)) +
        pp.Word("*", exact=1) +
        pp.Suppress(pp.Word(")", exact=1))
    )("function_action").
    setParseAction(lambda c: len)
)

# 'string'
# "string"
# "string, any type of string including commas"
quoted_var = pp.Or(
    [
        pp.QuotedString("'"),
        pp.QuotedString('"'),
    ]
)("alias")


# 'abc123' or "abc123" or abc123 or abc, 123
quote_unquote_var = pp.Or(
    [
        quoted_var,
        varnums("alias"),
    ]
)("alias")

# Single level dict parser, does not support nested dict expressions
# {}
# {'key': 'value'}
# {"key": "value"}
# {'key': value}
# {'key': "Value, value"}
# {'key': 'Value, value'}
# {'key', 'value', 'key', 'value', ....}
dict_literal = (
    (
        pp.Suppress(pp.Literal("{")) +
        pp.ZeroOrMore(
            quote_unquote_var +
            pp.Suppress(pp.Literal(":")) +
            quote_unquote_var +
            pp.Optional(pp.Suppress(pp.Literal(",")))
        ) +
        pp.Suppress(pp.Literal("}"))
    )("dict_literal")
).setParseAction(
    # convert a list into a dictionary
    # eg: ['name', 'Bob'] -> {'name': 'Bob'}
    lambda p: dict(zip(*[iter(p)]*2))
)


# One of more labels :LABEL_A or :LABEL_A:LABEL_B
labels = pp.ZeroOrMore(
    pp.Suppress(":") +
    pp.Group(varnums("label"))
)("labels")




# Open and close markers ()
vertex_open_marker = pp.Word("(", exact=1)("open_marker")
vertex_close_marker = pp.Word(")", exact=1)("close_marker")
edge_open_marker = pp.Word("[", exact=1)("open_marker")
edge_close_marker = pp.Word("]", exact=1)("close_marker")


###########################################
## Expressions for vertices/nodes        ##
###########################################

# Full vertex expression
# ()
# (var)
# (:label) or multi (:label1:label2) and so on
# ({'key': value})
# (var:label)
# (var {'key': value})
# (:label {'key': value})
# (var:label {'key': value})
vertex = pp.Group(
    pp.Suppress(vertex_open_marker) +
    pp.Optional(var("alias")) +
    pp.Optional(labels("labels")) +
    pp.Optional(dict_literal("properties")) +
    pp.Suppress(vertex_close_marker)
).setResultsName("vertex")


###########################################
## Expressions for edges/relationships   ##
###########################################

# *<start int>..<end int>
# *
# *..10
# *1..2
range_literal = pp.Group(
    pp.Suppress(pp.Word("*", exact=1)) +
    pp.Optional(
        pp.Or(
            [
                pp.Word("..", exact=2) + nums("max"),
                nums("min") + pp.Word("..", exact=2) + nums("max"),
            ]
        )
    )
)("range_literal")


# edge label
# :label
# :label|label
edge_label = pp.Forward()
edge_label << (
    labels +
    pp.ZeroOrMore(
        pp.Suppress(pp.Literal("|")) +
        edge_label
    )
)


# []
# [e]
# [:label] or multiple [:label1|label2]
# [*1..2]
# [{'key': value}]
# [e:label]
# [e *1..2]
# [e:label *]
# [e:label *1..2]
# [e:label *1..2 {'key': value}]
# [:label *]
# [:label *1..2]
# [:label {'key': value}]
# [*1..2 {'key': value}]
edge_body = (
    pp.Group(
        pp.Suppress(edge_open_marker) +
        pp.Optional(var("alias")) +
        pp.Optional(edge_label) +
        pp.Optional(range_literal) +
        pp.Optional(dict_literal) +
        pp.Suppress(edge_close_marker)
    )
)("edge_body")



# RelationshipPattern =
#   (LeftArrowHead, WS, Dash, WS, [RelationshipDetail], WS, Dash, WS, RightArrowHead) # pylint: disable=line-too-long
#   | (LeftArrowHead, WS, Dash, WS, [RelationshipDetail], WS, Dash)
#   | (Dash, WS, [RelationshipDetail], WS, Dash, WS, RightArrowHead)
#   | (Dash, WS, [RelationshipDetail], WS, Dash)
#   ;
edge = pp.Group(
    pp.Or(
        [
            # <-[]->
            (
                pp.Word("<-", exact=2)("in").setParseAction(lambda: True) +
                pp.Optional(edge_body) +
                pp.Word("->", exact=2)("out").setParseAction(lambda: True)
            ),

            # <-[]-
            (
                pp.Word("<-", exact=2)("in").setParseAction(lambda: True) +
                pp.Optional(edge_body) +
                pp.Word("-", exact=1)("out").setParseAction(lambda: False)
            ),

            # -[]->
            (
                pp.Word("-", exact=1)("in").setParseAction(lambda: False) +
                pp.Optional(edge_body) +
                pp.Word("->", exact=2)("out").setParseAction(lambda: True)
            ),

            # -[]-
            (
                pp.Word("-", exact=1)("in").setParseAction(lambda: True) +
                pp.Optional(edge_body) +
                pp.Word("-", exact=1)("out").setParseAction(lambda: True)
            ),
        ]
    )
)("edge")



# (n:Person)
# (n:Person),(m:Movie)
# (n:Person:Swedish)
# (n:Person {'name': 'Bob'})
# (n)-->(m)
# (n)--(m)
# (n:Person)-->(m)
# (n)<-[:KNOWS]-(m)
# (n)-[r]->(m)
# (n)-[*1..5]->(m)
# (n)-[:KNOWS]->(m:Person {'name': 'Bob'})
pattern = pp.Forward()
pattern << (
    vertex.setResultsName("vertices", listAllMatches=True) +
    pp.Optional(
        edge.setResultsName("edges", listAllMatches=True) +
        pattern
    )
) + pp.Optional(pp.Suppress(",") + pattern)


# m.key
property_key_name = pp.Group(
    varnums("alias") +
    pp.Suppress(pp.Literal(".")) +
    varnums("key")
)("property_key_name")


property_lookup = property_key_name("property_lookup")


###########################################
##              Expressions              ##
###########################################
expression = pp.Forward()

# Atom = NumberLiteral
#      | StringLiteral
#      | Parameter  --> not done
#      | (T,R,U,E)
#      | (F,A,L,S,E)
#      | (N,U,L,L)
#      | CaseExpression  --> not done
#      | ((C,O,U,N,T), '(', '*', ')')
#      | MapLiteral
#      | ListComprehension  --> not done
#      | ('[', WS, Expression, WS, { ',', WS, Expression, WS }, ']')
#      | ((F,I,L,T,E,R), WS, '(', WS, FilterExpression, WS, ')')  --> not done
#      | ((E,X,T,R,A,C,T), WS, '(', WS, FilterExpression, WS, [WS, '|', Expression], ')')  --> not done
#      | Reduce  --> not done
#      | ((A,L,L), WS, '(', WS, FilterExpression, WS, ')') --> not done
#      | ((A,N,Y), WS, '(', WS, FilterExpression, WS, ')') --> not done
#      | ((N,O,N,E), WS, '(', WS, FilterExpression, WS, ')') --> not done
#      | ((S,I,N,G,L,E), WS, '(', WS, FilterExpression, WS, ')') --> not done
#      | ShortestPathPattern --> not done
#      | RelationshipsPattern --> not done
#      | parenthesizedExpression --> not done
#      | FunctionInvocation --> not done
#      | Variable
#      ;
expression_atom = pp.Or(
    [
        # NULL
        null_keyword("null"),

        # TRUE
        true_keyword,

        # FALSE
        false_keyword,

        # COUNT (*)
        count,

        # 1234
        nums,

        # abcd
        var,

        # {'key': 'value'}
        dict_literal("dict"),

        # [expr]
        # [expr, expr]
        (
            pp.Word("[", exact=1) +
            expression +
            pp.ZeroOrMore(
                pp.Word(",", exact=1) + expression
            ) +
            pp.Word("]", exact=1)
        )
    ]
)


# Expression2 = Atom, { PropertyLookup | NodeLabels } ;
expression2 = (
    expression_atom + pp.ZeroOrMore(
        pp.Or(
            [
                property_lookup,
                labels,
            ]
        )
    )
)

# Expression3 = Expression2, {
#   (WS, '[', Expression, ']') |
#   (WS, '[', [Expression], '..', [Expression], ']') |
#   (((WS, '=~') | (SP, (I,N)) | (SP, (S,T,A,R,T,S), SP, (W,I,T,H)) | (SP, (E,N,D,S), SP, (W,I,T,H)) | (SP, (C,O,N,T,A,I,N,S))), WS, Expression2) |
#   (SP, (I,S), SP, (N,U,L,L)) |
#   (SP, (I,S), SP, (N,O,T), SP, (N,U,L,L))
#} ;
expression3 = (
    expression2 +
    pp.ZeroOrMore(
        pp.Or(
            [
                #   (WS, '[', Expression, ']') |
                pp.Word("[", exact=1) + expression + pp.Word("]", exact=1),

                #   (WS, '[', [Expression], '..', [Expression], ']') |
                (
                    pp.Word("[", exact=1) +
                    (
                        pp.Optional(expression) +
                        pp.Word("..", exact=2) +
                        pp.Optional(expression)
                    ) +
                    pp.Word("]", exact=1)
                ),

                #   (((WS, '=~') | (SP, (I,N)) | (SP, (S,T,A,R,T,S), SP, (W,I,T,H)) | (SP, (E,N,D,S), SP, (W,I,T,H)) | (SP, (C,O,N,T,A,I,N,S))), WS, Expression2) |
                (
                    pp.Or(
                        [
                            regex_operation,
                            in_keyword,
                            starts_with,
                            ends_with,
                            contains,
                        ]
                    ) + expression2
                ),

                #   (SP, (I,S), SP, (N,U,L,L)) |
                is_null,

                #   (SP, (I,S), SP, (N,O,T), SP, (N,U,L,L))
                is_not_null,
            ]
        )
    )
)

# Expression4 = { ('+' | '-'), WS }, Expression3 ;
expression4 = (
    pp.ZeroOrMore(
        pp.Or(
            [
                add_operation,
                subtract_operation,
            ]
        )
    ) + expression3
)

# Expression5 = Expression4, { WS, '^', WS, Expression4 } ;
expression5 = (
    expression4 + pp.ZeroOrMore(bitwise_or_operation + expression4)
)

# Expression6 = Expression5, { (WS, '*', WS, Expression5) | (WS, '/', WS, Expression5) | (WS, '%', WS, Expression5) } ;
expression6 = (
    expression5 + pp.ZeroOrMore(
        pp.Or(
            [
                multiply_operation + expression5,
                divide_operation + expression5,
                percent_operation + expression5,
            ]
        )
    )
)

# Expression7 = Expression6, { (WS, '+', WS, Expression6) | (WS, '-', WS, Expression6) } ;
expression7 = (
    expression6 + pp.ZeroOrMore(
        pp.Or(
            [
                add_operation + expression6,
                subtract_operation + expression6,
            ]
        )
    )
)

# PartialComparisonExpression = ('=', WS, Expression7)
#                             | ('<>', WS, Expression7)
#                             | ('!=', WS, Expression7)
#                             | ('<', WS, Expression7)
#                             | ('>', WS, Expression7)
#                             | ('<=', WS, Expression7)
#                             | ('>=', WS, Expression7)
#                             ;
partial_comparison_expression = (
    pp.Or(
        [
            eq_operation + expression7,
            neq_operation + expression7,
            lt_operation + expression7,
            lte_operation + expression7,
            gt_operation + expression7,
            gte_operation + expression7,
        ]
    )
)


# Expression8 = Expression7, { WS, PartialComparisonExpression } ;
expression8 = (
    expression7 + pp.ZeroOrMore(partial_comparison_expression)
)


# Expression9 = { SP, (N,O,T), SP }, Expression8 ;
expression9 = (pp.ZeroOrMore(pp.CaselessKeyword("NOT")) + expression8)


# Expression10 = Expression9, { SP, (A,N,D), SP, Expression9 } ;
expression10 = (
    expression9 + pp.ZeroOrMore(
        pp.CaselessKeyword("AND") + expression9
    )
)


# Expression11 = Expression10, { SP, (X,O,R), SP, Expression10 } ;
expression11 = (
    expression10 + pp.ZeroOrMore(
        pp.CaselessKeyword("XOR") + expression10
    )
)


# Expression12 = Expression11, { SP, (O,R), SP, Expression11 } ;
expression12 = (
    expression11 + pp.ZeroOrMore(
        pp.CaselessKeyword("OR") + expression11
    )
)


# Expression = Expression12 ;
expression << expression12  # pylint: disable=pointless-statement


###########################################
##              Clauses                  ##
###########################################

# MATCH
# Match
# match
match_clause = pp.CaselessKeyword("MATCH")("clause")


# WHERE
# Where
# where
where_clause = pp.CaselessKeyword("WHERE")("clause")


# AS
# As
# as
as_clause = pp.CaselessKeyword("AS")("clause")


# RETURN
# Return
# return
return_clause = pp.CaselessKeyword("RETURN")("clause")


# WHERE PATTERN
# WHERE m.key = value
# WHERE m.key != value
# WHERE m.key > value
# WHERE m.key >= value
# WHERE m.key < value
# WHERE m.key <= value
where_pattern = pp.Group(
    where_clause +
    property_lookup +
    pp.Or(
        [
            eq_operation,
            neq_operation,
            lt_operation,
            lte_operation,
            gt_operation,
            gte_operation,
        ]
    ) +
    quote_unquote_var("value")
)("where_pattern")


# MATCH PATTERN
# MATCH (n)-[]-(m)
# MATCH (n)-[]-(m) WHERE n.key = value
# MATCH (n)-[]-(m)<-[]-() WHERE n.key = value
match_pattern = pp.Group(
    match_clause +
    pattern +
    pp.Optional(where_pattern)
)("match_pattern")


# DISTINCT
# Distinct
# distinct
# distinct_action = (
#     pp.CaselessKeyword("DISTINCT").
#     setResultsName("action").
#     setParseAction(lambda t: "distinct")
# )


# UNION
# Union
# union
# union_action = (
#     pp.CaselessKeyword("UNION").
#     setResultsName("action").
#     setParseAction(lambda t: "union")
# )



# TODO:
# This is a hack till I figure out the serious performance issues with the
# above. return_item should be using expression and not expression_atom.

# ReturnItem = (Expression, SP, (A,S), SP, Variable)| Expression;
return_item = pp.Group(
    pp.Or(
        [
            expression_atom + as_clause + var,
            expression_atom,
        ]
    )
)("return_item")


# ReturnItems = ('*', { WS, ',', WS, ReturnItem })|
#               (ReturnItem, { WS, ',', WS, ReturnItem });
return_items = pp.Group(
    pp.Or(
        [
            pp.Word("*", exact=1) +
            pp.ZeroOrMore(
                pp.Word(",", exact=1) + return_item
            ),

            return_item + pp.ZeroOrMore(
                pp.Word(",", exact=1) + return_item
            ),
        ]
    )
)("return_items")


# ReturnBody = ReturnItems, [SP, Order], [SP, Skip], [SP, Limit] ;
return_body = (
    return_items # +
    # pp,Optional(
    #     order_action | skip_action | limit_action
    # )
)


# Return = ((R,E,T,U,R,N), SP, (D,I,S,T,I,N,C,T), SP, ReturnBody)|
#          ((R,E,T,U,R,N), SP, ReturnBody);
return_pattern = (
    return_clause + return_body
)("return_pattern")


# only supporting MATCH for now but should also support CREATE in the near
# future
clause = (
    match_pattern
)("clause")


# SingleQuery = Clause, { WS, Clause } ;
single_query = (
    clause +
    pp.ZeroOrMore(clause)
)("single_query")


# not supporting Unions at this stage
# Union = ((U,N,I,O,N), SP, (A,L,L), SingleQuery) | ((U,N,I,O,N), SingleQuery)
# union = (
#     union_action, pp.CaselessKeyword("ALL") + single_query |
#     union_action + single_query
# )


# RegularQuery = SingleQuery, { WS, Union } ;
regular_query = (
    single_query
    # + pp.ZeroOrMore(union)
)("regular_query")


# not supporting BulkImportQuery at this stage.
# Query = RegularQuery | BulkImportQuery;
query = (regular_query + return_pattern)("query")


def parse(query_str, expr=query, string_end=True, enable_packrat=True):
    """
    Parse the given query string using the provided grammar expression.

    :param query_str: Query string that you are parsing.
    :type query_str: :class:`str`
    :param expr: Grammar expression used for parsing the query string.
    :type expr: :class:`pyparsing.ParserElement`
    :param string_end: True to parsing the entire query string till the
        very end.
    :type string_end: :class:`bool`
    :param enable_packrat: Packrat is a grammar traversal algorithm. It will
        speed things up nicely. This option is not turned on by default with
        pyparsing.
    :type enable_packrat: :class:`bool`
    :returns: Dictionary of parsing results extracted by each grammar
        expression.
    :rtype: :class:`dict`
    """
    if string_end is True:
        expr = expr + pp.StringEnd()

    if enable_packrat:
        expr.enablePackrat()

    return expr.parseString(query_str).asDict()



#pylint: skip-file

import logging
import inspect
import ruruki

from ruruki.graphs import Graph


GRAPH = Graph()
GRAPH.add_vertex_constraint("class", "name")
GRAPH.add_vertex_constraint("method", "name")
GRAPH.add_vertex_constraint("file", "name")
GRAPH.add_vertex_constraint("function", "name")
GRAPH.add_vertex_constraint("module", "name")


SEEN = set()

def build_dep(lib, parent):
    previous = parent

    for name, module in inspect.getmembers(lib, inspect.ismodule):
        if module in SEEN:
            continue
        SEEN.add(module)
        parent = GRAPH.get_or_create_vertex("module", name=name)

        # link to the previous parent
        GRAPH.get_or_create_edge(parent, "comes-form", previous)

        try:
            filename = inspect.getfile(module)
            if filename:
                parent_file = GRAPH.get_or_create_vertex("file", name=filename)
                GRAPH.get_or_create_edge(parent, "found-in", parent_file)
        except TypeError:
            logging.warn("Failed to get the file for %s", name)


        # get all the functions in the module
        for fname, func in inspect.getmembers(module, inspect.isfunction):
            fnode = GRAPH.get_or_create_vertex("function", name=fname)
            GRAPH.get_or_create_edge(parent, "has-function", fnode)

        # get all the classes in the module
        for cname, cls in inspect.getmembers(module, inspect.isclass):
            cnode = GRAPH.get_or_create_vertex("class", name=cname)
            GRAPH.get_or_create_edge(parent, "has-class", cnode)

            # get all the class methods in the class
            for mname, meth in inspect.getmembers(cls, inspect.ismethod):
                mnode = GRAPH.get_or_create_vertex("method", name=mname)
                GRAPH.get_or_create_edge(cnode, "has-method", mnode)

        build_dep(module, parent)


def scrape():
    build_dep(ruruki, GRAPH.get_or_create_vertex("package", name="ruruki"))
    return GRAPH


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    from ruruki_eye.server import run

    scrape()
    run("0.0.0.0", 8000, False, GRAPH)
